/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */
// Word cloud layout by Jason Davies, https://www.jasondavies.com/wordcloud/ https://github.com/jasondavies/d3-cloud
// Algorithm due to Jonathan Feinberg, http://static.mrfeinberg.com/bv_ch03.pdf

var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var WordCloud1447959067750;
        (function (WordCloud1447959067750) {
            var ValueFormatter = powerbi.visuals.valueFormatter;
            var getAnimationDuration = visuals.AnimatorCommon.GetAnimationDuration;
            (function (WordCloudScaleType) {
                WordCloudScaleType[WordCloudScaleType["logn"] = 0] = "logn";
                WordCloudScaleType[WordCloudScaleType["sqrt"] = 1] = "sqrt";
                WordCloudScaleType[WordCloudScaleType["value"] = 2] = "value";
            })(WordCloud1447959067750.WordCloudScaleType || (WordCloud1447959067750.WordCloudScaleType = {}));
            var WordCloudScaleType = WordCloud1447959067750.WordCloudScaleType;
            ;
            var ValueSelectionManager = (function () {
                function ValueSelectionManager(hostServices, getSelectionIds) {
                    this.selectedValuesValue = [];
                    this.hostServices = hostServices;
                    this.selectionManager = new visuals.utility.SelectionManager({ hostServices: hostServices });
                    this.getSelectionIds = function (value) { return _.isArray(value)
                        ? _.flatten(value.map(function (x) { return getSelectionIds(x); }))
                        : getSelectionIds(value); };
                }
                Object.defineProperty(ValueSelectionManager.prototype, "selectedValues", {
                    get: function () {
                        return this.selectedValuesValue;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(ValueSelectionManager.prototype, "selectionIds", {
                    get: function () {
                        return this.getSelectionIds(this.selectedValues);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(ValueSelectionManager.prototype, "hasSelection", {
                    get: function () {
                        return this.selectedValues.length > 0;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(ValueSelectionManager.prototype, "getSelectionIdValues", {
                    get: function () {
                        var _this = this;
                        return this.selectedValues.map(function (v) { return { value: v, selectionId: _this.getSelectionIds(v) }; });
                    },
                    enumerable: true,
                    configurable: true
                });
                ValueSelectionManager.prototype.selectAndSendSelection = function (value, multiSelect) {
                    if (multiSelect === void 0) { multiSelect = false; }
                    var values = (_.isArray(value) ? value : [value]);
                    if (this.hostServices.shouldRetainSelection()) {
                        return this.sendSelectionToHost(this.getSelectionIds(values));
                    }
                    else {
                        this.selectInternal(values, multiSelect);
                        return this.sendSelection();
                    }
                };
                ValueSelectionManager.prototype.select = function (value, multiSelect) {
                    if (multiSelect === void 0) { multiSelect = false; }
                    var values = (_.isArray(value) ? value : [value]);
                    this.selectInternal(values, multiSelect);
                };
                ValueSelectionManager.prototype.isSelected = function (selectionId) {
                    var _this = this;
                    var values = (_.isArray(selectionId) ? selectionId : [selectionId]);
                    return values.every(function (v) { return _this.selectedValues.some(function (s) { return s === v; }); });
                };
                ValueSelectionManager.prototype.sendSelection = function () {
                    return this.sendSelectionToHost(this.selectionIds);
                };
                ValueSelectionManager.prototype.clear = function (sendToHost) {
                    this.selectedValues.length = 0;
                    if (sendToHost) {
                        return this.sendSelection();
                    }
                    return $.Deferred().resolve();
                };
                ValueSelectionManager.prototype.selectInternal = function (values, multiSelect) {
                    var _this = this;
                    var resultValues = [];
                    if (this.isSelected(values)) {
                        resultValues = multiSelect
                            ? this.selectedValues.filter(function (s) { return !values.some(function (v) { return v === s; }); })
                            : this.selectedValues.length === values.length ? [] : values;
                    }
                    else {
                        resultValues = multiSelect
                            ? values.filter(function (x) { return !_this.isSelected(x); }).concat(this.selectedValues)
                            : values;
                    }
                    this.selectedValues.length = 0;
                    resultValues.forEach(function (x) { return _this.selectedValues.push(x); });
                };
                ValueSelectionManager.prototype.sendSelectionToHost = function (ids) {
                    var deferred = $.Deferred();
                    this.selectionManager.sendSelectionToHost(ids);
                    deferred.resolve(this.selectionIds);
                    return deferred;
                };
                return ValueSelectionManager;
            }());
            var VisualLayout = (function () {
                function VisualLayout(defaultViewport, defaultMargin) {
                    this.defaultViewport = defaultViewport || { width: 0, height: 0 };
                    this.defaultMargin = defaultMargin || { top: 0, bottom: 0, right: 0, left: 0 };
                }
                Object.defineProperty(VisualLayout.prototype, "viewport", {
                    get: function () {
                        return this.viewportValue || (this.viewportValue = this.defaultViewport);
                    },
                    set: function (value) {
                        var _this = this;
                        this.previousOriginalViewportValue = _.clone(this.originalViewportValue);
                        this.originalViewportValue = _.clone(value);
                        this.setUpdateObject(value, function (v) { return _this.viewportValue = v; }, function (o) { return VisualLayout.restrictToMinMax(o, _this.minViewport); });
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "viewportCopy", {
                    get: function () {
                        return _.clone(this.viewport);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "viewportIn", {
                    //Returns viewport minus margin
                    get: function () {
                        return this.viewportInValue || this.viewport;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "minViewport", {
                    get: function () {
                        return this.minViewportValue || { width: 0, height: 0 };
                    },
                    set: function (value) {
                        var _this = this;
                        this.setUpdateObject(value, function (v) { return _this.minViewportValue = v; }, VisualLayout.restrictToMinMax);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "margin", {
                    get: function () {
                        return this.marginValue || (this.marginValue = this.defaultMargin);
                    },
                    set: function (value) {
                        var _this = this;
                        this.setUpdateObject(value, function (v) { return _this.marginValue = v; }, VisualLayout.restrictToMinMax);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "viewportChanged", {
                    //Returns true if viewport has updated after last change.
                    get: function () {
                        return !!this.originalViewportValue && (!this.previousOriginalViewportValue
                            || this.previousOriginalViewportValue.height !== this.originalViewportValue.height
                            || this.previousOriginalViewportValue.width !== this.originalViewportValue.width);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(VisualLayout.prototype, "viewportInIsZero", {
                    get: function () {
                        return this.viewportIn.width === 0 || this.viewportIn.height === 0;
                    },
                    enumerable: true,
                    configurable: true
                });
                VisualLayout.prototype.resetMargin = function () {
                    this.margin = this.defaultMargin;
                };
                VisualLayout.prototype.update = function () {
                    this.viewportInValue = VisualLayout.restrictToMinMax({
                        width: this.viewport.width - (this.margin.left + this.margin.right),
                        height: this.viewport.height - (this.margin.top + this.margin.bottom)
                    }, this.minViewportValue);
                };
                VisualLayout.prototype.setUpdateObject = function (object, setObjectFn, beforeUpdateFn) {
                    var _this = this;
                    object = _.clone(object);
                    setObjectFn(VisualLayout.createNotifyChangedObject(object, function (o) {
                        if (beforeUpdateFn)
                            beforeUpdateFn(object);
                        _this.update();
                    }));
                    if (beforeUpdateFn)
                        beforeUpdateFn(object);
                    this.update();
                };
                VisualLayout.createNotifyChangedObject = function (object, objectChanged) {
                    var result = {};
                    _.keys(object).forEach(function (key) { return Object.defineProperty(result, key, {
                        get: function () { return object[key]; },
                        set: function (value) { object[key] = value; objectChanged(object, key); },
                        enumerable: true,
                        configurable: true
                    }); });
                    return result;
                };
                VisualLayout.restrictToMinMax = function (value, minValue) {
                    _.keys(value).forEach(function (x) { return value[x] = Math.max(minValue && minValue[x] || 0, value[x]); });
                    return value;
                };
                return VisualLayout;
            }());
            var WordCloudSettings = (function () {
                function WordCloudSettings() {
                    //Default Settings
                    this.general = {
                        maxNumberOfWords: 200,
                        minFontSize: 20 / WordCloud.FontSizePercentageCoefficent,
                        maxFontSize: 100 / WordCloud.FontSizePercentageCoefficent,
                        isBrokenText: true
                    };
                    this.stopWords = {
                        show: true,
                        isDefaultStopWords: false,
                        words: null
                    };
                    this.rotateText = {
                        show: true,
                        minAngle: -60,
                        maxAngle: 90,
                        maxNumberOfOrientations: 2
                    };
                }
                Object.defineProperty(WordCloudSettings, "Default", {
                    get: function () {
                        return new this();
                    },
                    enumerable: true,
                    configurable: true
                });
                WordCloudSettings.parse = function (dataView, capabilities) {
                    var settings = new this();
                    if (!dataView || !dataView.metadata || !dataView.metadata.objects) {
                        return settings;
                    }
                    var properties = this.getProperties(capabilities);
                    for (var objectKey in capabilities.objects) {
                        for (var propKey in capabilities.objects[objectKey].properties) {
                            if (!settings[objectKey] || !_.has(settings[objectKey], propKey)) {
                                continue;
                            }
                            var type = capabilities.objects[objectKey].properties[propKey].type;
                            var getValueFn = this.getValueFnByType(type);
                            settings[objectKey][propKey] = getValueFn(dataView.metadata.objects, properties[objectKey][propKey], settings[objectKey][propKey]);
                        }
                    }
                    return settings;
                };
                WordCloudSettings.getProperties = function (capabilities) {
                    var objects = _.merge({
                        general: { properties: { formatString: {} } }
                    }, capabilities.objects);
                    var properties = {};
                    for (var objectKey in objects) {
                        properties[objectKey] = {};
                        for (var propKey in objects[objectKey].properties) {
                            properties[objectKey][propKey] = {
                                objectName: objectKey,
                                propertyName: propKey
                            };
                        }
                    }
                    return properties;
                };
                WordCloudSettings.createEnumTypeFromEnum = function (type) {
                    var even = false;
                    return powerbi.createEnumType(Object.keys(type)
                        .filter(function (key, i) { return ((!!(i % 2)) === even && type[key] === key
                        && !void (even = !even)) || (!!(i % 2)) !== even; })
                        .map(function (x) { return { value: x, displayName: x }; }));
                };
                WordCloudSettings.getValueFnByType = function (type) {
                    switch (_.keys(type)[0]) {
                        case "fill":
                            return powerbi.DataViewObjects.getFillColor;
                        default:
                            return powerbi.DataViewObjects.getValue;
                    }
                };
                WordCloudSettings.enumerateObjectInstances = function (settings, options, capabilities) {
                    if (settings === void 0) { settings = new this(); }
                    var enumeration = new visuals.ObjectEnumerationBuilder();
                    var object = settings && settings[options.objectName];
                    if (!object) {
                        return enumeration;
                    }
                    var instance = {
                        objectName: options.objectName,
                        selector: null,
                        properties: {}
                    };
                    for (var key in object) {
                        if (_.has(object, key)) {
                            instance.properties[key] = object[key];
                        }
                    }
                    enumeration.pushInstance(instance);
                    return enumeration;
                };
                WordCloudSettings.prototype.createOriginalSettings = function () {
                    this.originalSettings = _.cloneDeep(this);
                };
                return WordCloudSettings;
            }());
            WordCloud1447959067750.WordCloudSettings = WordCloudSettings;
            var WordCloudColumns = (function () {
                function WordCloudColumns() {
                    //Data Roles
                    this.Category = null;
                    this.Values = null;
                }
                WordCloudColumns.getColumnSources = function (dataView) {
                    return this.getColumnSourcesT(dataView);
                };
                WordCloudColumns.getTableValues = function (dataView) {
                    var table = dataView && dataView.table;
                    var columns = this.getColumnSourcesT(dataView);
                    return columns && table && _.mapValues(columns, function (n, i) { return n && table.rows.map(function (row) { return row[n.index]; }); });
                };
                WordCloudColumns.getTableRows = function (dataView) {
                    var table = dataView && dataView.table;
                    var columns = this.getColumnSourcesT(dataView);
                    return columns && table && table.rows.map(function (row) {
                        return _.mapValues(columns, function (n, i) { return n && row[n.index]; });
                    });
                };
                WordCloudColumns.getCategoricalValues = function (dataView) {
                    var categorical = dataView && dataView.categorical;
                    var categories = categorical && categorical.categories || [];
                    var values = categorical && categorical.values || [];
                    var series = categorical && values.source && this.getSeriesValues(dataView);
                    return categorical && _.mapValues(new this(), function (n, i) {
                        return _.toArray(categories).concat(_.toArray(values))
                            .filter(function (x) { return x.source.roles && x.source.roles[i]; }).map(function (x) { return x.values; })[0]
                            || values.source && values.source.roles && values.source.roles[i] && series;
                    });
                };
                WordCloudColumns.getSeriesValues = function (dataView) {
                    return dataView && dataView.categorical && dataView.categorical.values
                        && dataView.categorical.values.map(function (x) { return visuals.converterHelper.getSeriesName(x.source); });
                };
                WordCloudColumns.getCategoricalColumns = function (dataView) {
                    var categorical = dataView && dataView.categorical;
                    var categories = categorical && categorical.categories || [];
                    var values = categorical && categorical.values || [];
                    return categorical && _.mapValues(new this(), function (n, i) { return categories.filter(function (x) { return x.source.roles && x.source.roles[i]; })[0]
                        || values.source && values.source.roles && values.source.roles[i] && values
                        || values.filter(function (x) { return x.source.roles && x.source.roles[i]; }); });
                };
                WordCloudColumns.getColumnSourcesT = function (dataView) {
                    var columns = dataView && dataView.metadata && dataView.metadata.columns;
                    return columns && _.mapValues(new this(), function (n, i) { return columns.filter(function (x) { return x.roles && x.roles[i]; })[0]; });
                };
                WordCloudColumns.Roles = Object.freeze(_.mapValues(new WordCloudColumns(), function (x, i) { return i; }));
                return WordCloudColumns;
            }());
            WordCloud1447959067750.WordCloudColumns = WordCloudColumns;
            var WordCloud = (function () {
                function WordCloud(options) {
                    this.durationAnimations = 500;
                    this.fakeViewport = {
                        width: 1500,
                        height: 1000
                    };
                    this.canvasViewport = {
                        width: 128,
                        height: 2048
                    };
                    if (options) {
                        this.svg = options.svg || this.svg;
                        this.layout = new VisualLayout(null, options.margin || WordCloud.DefaultMargin);
                        if (options.animator)
                            this.animator = options.animator;
                    }
                    this.isUpdating = false;
                }
                WordCloud.converter = function (dataView, colors, previousData) {
                    var categorical = WordCloudColumns.getCategoricalColumns(dataView);
                    if (!categorical || !categorical.Category || _.isEmpty(categorical.Category.values)) {
                        return null;
                    }
                    var catValues = WordCloudColumns.getCategoricalValues(dataView);
                    var properties = WordCloudSettings.getProperties(WordCloud.capabilities);
                    var settings = WordCloud.parseSettings(dataView, previousData && previousData.settings);
                    var wordValueFormatter = ValueFormatter.create({
                        format: ValueFormatter.getFormatString(categorical.Category.source, properties.general.formatString),
                        value: catValues.Category[0]
                    });
                    var stopWords = _.isString(settings.stopWords.words) ? settings.stopWords.words.split(WordCloud.StopWordsDelemiter) : [];
                    stopWords = settings.stopWords.isDefaultStopWords ? stopWords.concat(WordCloud.StopWords) : stopWords;
                    var colorHelper = new visuals.ColorHelper(colors, properties.dataPoint.fill, explore.util.getRandomColor());
                    var texts = catValues.Category.map(function (item, index) {
                        var color;
                        if (categorical.Category.objects && categorical.Category.objects[index]) {
                            color = explore.util.hexToRgb(colorHelper.getColorForMeasure(categorical.Category.objects[index], ""));
                        }
                        else {
                            color = previousData && previousData.texts && previousData.texts[index]
                                ? previousData.texts[index].color
                                : explore.util.getRandomColor();
                        }
                        var selectionId = new visuals.SelectionIdBuilder()
                            .withCategory(dataView.categorical.categories[0], index)
                            .createSelectionId();
                        return {
                            text: item,
                            count: (catValues.Values && catValues.Values[index] && !isNaN(catValues.Values[index])) ? catValues.Values[index] : 1,
                            index: index,
                            selectionId: selectionId,
                            color: color,
                            textGroup: item
                        };
                    });
                    var reducedTexts = WordCloud.getReducedText(texts, stopWords, settings);
                    var dataPoints = WordCloud.getDataPoints(reducedTexts, settings, wordValueFormatter);
                    return {
                        dataView: dataView,
                        settings: settings,
                        texts: texts,
                        dataPoints: dataPoints
                    };
                };
                WordCloud.parseSettings = function (dataView, previousSettings) {
                    var settings = WordCloudSettings.parse(dataView, WordCloud.capabilities);
                    settings.general.minFontSize = Math.max(settings.general.minFontSize, 1);
                    settings.general.maxFontSize = Math.max(settings.general.maxFontSize, 1);
                    settings.general.maxFontSize = Math.max(settings.general.maxFontSize, settings.general.minFontSize);
                    settings.rotateText.minAngle = Math.max(Math.min(settings.rotateText.minAngle, WordCloud.MaxAngle), WordCloud.MinAngle);
                    settings.rotateText.maxAngle = Math.max(Math.min(settings.rotateText.maxAngle, WordCloud.MaxAngle), WordCloud.MinAngle);
                    settings.rotateText.maxAngle = Math.max(settings.rotateText.maxAngle, settings.rotateText.minAngle);
                    settings.general.maxNumberOfWords = Math.max(Math.min(settings.general.maxNumberOfWords, WordCloud.MaxNumberOfWords), 1);
                    settings.rotateText.maxNumberOfOrientations = Math.max(Math.min(settings.rotateText.maxNumberOfOrientations, WordCloud.MaxNumberOfWords), 1);
                    settings.createOriginalSettings();
                    return settings;
                };
                WordCloud.getReducedText = function (texts, stopWords, settings) {
                    var brokenStrings = WordCloud.getBrokenWords(texts, stopWords, settings);
                    var result = _.values(_.groupBy(brokenStrings, function (x) { return x.text; }));
                    result = result.map(function (texts) { return _.sortBy(texts, function (x) { return x.textGroup.length; }); });
                    return result;
                };
                WordCloud.getBrokenWords = function (words, stopWords, settings) {
                    var brokenStrings = [];
                    var whiteSpaceRegExp = /\s/;
                    var punctuatuinRegExp = new RegExp("[" + WordCloud.Punctuation.join("\\") + "]", "gim");
                    if (!settings.general.isBrokenText) {
                        return words;
                    }
                    words.forEach(function (item) {
                        if (typeof item.text === "string") {
                            var splittedWords = item.text.replace(punctuatuinRegExp, " ").split(whiteSpaceRegExp);
                            if (settings.stopWords.show) {
                                splittedWords = splittedWords.filter(function (value) {
                                    return value.length > 0 && !stopWords.some(function (removeWord) {
                                        return value.toLocaleLowerCase() === removeWord.toLocaleLowerCase();
                                    });
                                });
                            }
                            splittedWords.forEach(function (splittedWord) {
                                if (splittedWord.length === 0 || whiteSpaceRegExp.test(splittedWord)) {
                                    return;
                                }
                                var existingWord = _.find(words, function (w) { return w.text.toString().toLocaleLowerCase() === splittedWord.toLocaleLowerCase(); });
                                if (existingWord && existingWord !== item) {
                                    splittedWord = existingWord.text;
                                }
                                brokenStrings.push({
                                    text: splittedWord,
                                    textGroup: item.textGroup,
                                    count: item.count,
                                    index: item.index,
                                    selectionId: item.selectionId,
                                    color: item.color
                                });
                            });
                        }
                        else {
                            brokenStrings.push(item);
                        }
                    });
                    return brokenStrings;
                };
                WordCloud.getDataPoints = function (textGroups, settings, wordValueFormatter) {
                    if (_.isEmpty(textGroups)) {
                        return [];
                    }
                    var returnValues = textGroups.map(function (values) {
                        return {
                            text: wordValueFormatter.format(values[0].text),
                            x: 0,
                            y: 0,
                            rotate: WordCloud.getAngle(settings),
                            padding: 1,
                            width: 0,
                            height: 0,
                            xOff: 0,
                            yOff: 0,
                            x0: 0,
                            y0: 0,
                            x1: 0,
                            y1: 0,
                            color: values[0].color,
                            selectionIds: values.map(function (x) { return x.selectionId; }),
                            wordIndex: values[0].index,
                            count: _.sum(values, function (x) { return x.count; })
                        };
                    });
                    var minValue = _.min(returnValues, function (x) { return x.count; }).count;
                    var maxValue = _.max(returnValues, function (x) { return x.count; }).count;
                    var texts = textGroups.map(function (x) { return x[0]; });
                    returnValues.forEach(function (x) { return x.size = WordCloud.getWordFontSize(texts, settings, x.count, minValue, maxValue); });
                    return returnValues.sort(function (a, b) { return b.count - a.count; });
                };
                WordCloud.getWordFontSize = function (texts, settings, value, minValue, maxValue, scaleType) {
                    if (scaleType === void 0) { scaleType = WordCloudScaleType.value; }
                    var weight, fontSize;
                    var minFontSize = settings.general.minFontSize * WordCloud.FontSizePercentageCoefficent;
                    var maxFontSize = settings.general.maxFontSize * WordCloud.FontSizePercentageCoefficent;
                    if (texts.length < 2) {
                        return maxFontSize;
                    }
                    switch (scaleType) {
                        case WordCloudScaleType.logn: {
                            weight = Math.log(value);
                        }
                        case WordCloudScaleType.sqrt: {
                            weight = Math.sqrt(value);
                        }
                        case WordCloudScaleType.value: {
                            weight = value;
                        }
                    }
                    if (weight > minValue) {
                        fontSize = (maxValue - minValue) !== 0
                            ? (maxFontSize * (weight - minValue)) / (maxValue - minValue)
                            : 0;
                    }
                    else {
                        fontSize = 0;
                    }
                    fontSize = (fontSize * 100) / maxFontSize;
                    fontSize = (fontSize * (maxFontSize - minFontSize)) / 100 + minFontSize;
                    return fontSize;
                };
                WordCloud.getAngle = function (settings) {
                    if (!settings.rotateText.show) {
                        return 0;
                    }
                    var angle = ((settings.rotateText.maxAngle - settings.rotateText.minAngle)
                        / settings.rotateText.maxNumberOfOrientations)
                        * Math.floor(Math.random() * settings.rotateText.maxNumberOfOrientations);
                    return settings.rotateText.minAngle + angle;
                };
                Object.defineProperty(WordCloud.prototype, "settings", {
                    get: function () {
                        return this.data && this.data.settings;
                    },
                    enumerable: true,
                    configurable: true
                });
                WordCloud.prototype.init = function (options) {
                    var _this = this;
                    if (this.svg) {
                        this.root = this.svg;
                    }
                    else {
                        this.root = d3.select(options.element.get(0)).append("svg");
                    }
                    this.colors = options.style.colorPalette.dataColors;
                    this.hostService = options.host;
                    this.selectionManager = new ValueSelectionManager(this.hostService, function (w) {
                        var dataPoints = _this.data && _this.data.dataPoints && _this.data.dataPoints.filter(function (x) { return x.text === w; });
                        return dataPoints && dataPoints[0] && dataPoints[0].selectionIds;
                    });
                    this.layout = new VisualLayout(null, WordCloud.DefaultMargin);
                    this.root.classed(WordCloud.ClassName, true);
                    this.root.on("click", function () {
                        _this.setSelection(null);
                    });
                    this.fontFamily = this.root.style("font-family");
                    this.main = this.root.append("g");
                    this.wordsContainerSelection = this.main
                        .append("g")
                        .classed(WordCloud.Words["class"], true);
                    this.canvas = document.createElement("canvas");
                };
                WordCloud.prototype.update = function (visualUpdateOptions) {
                    var _this = this;
                    if (!visualUpdateOptions ||
                        !visualUpdateOptions.viewport ||
                        !visualUpdateOptions.dataViews ||
                        !visualUpdateOptions.dataViews[0] ||
                        !visualUpdateOptions.viewport ||
                        !(visualUpdateOptions.viewport.height >= 0) ||
                        !(visualUpdateOptions.viewport.width >= 0))
                        return;
                    if (visualUpdateOptions !== this.visualUpdateOptions) {
                        this.incomingUpdateOptions = visualUpdateOptions;
                    }
                    if (!this.isUpdating && (this.incomingUpdateOptions !== this.visualUpdateOptions)) {
                        this.visualUpdateOptions = this.incomingUpdateOptions;
                        this.layout.viewport = this.visualUpdateOptions.viewport;
                        var dataView = visualUpdateOptions.dataViews[0];
                        if (this.layout.viewportInIsZero) {
                            return;
                        }
                        this.durationAnimations = getAnimationDuration(this.animator, visualUpdateOptions.suppressAnimations);
                        this.UpdateSize();
                        var data = WordCloud.converter(dataView, this.colors, this.data);
                        if (!data) {
                            this.clear();
                            return;
                        }
                        this.data = data;
                        this.computePositions(function (wordCloudDataView) { return _this.render(wordCloudDataView); });
                    }
                };
                WordCloud.prototype.clear = function () {
                    this.main
                        .select(WordCloud.Words.selector)
                        .selectAll(WordCloud.WordGroup.selector)
                        .remove();
                };
                WordCloud.prototype.computePositions = function (onPositionsComputed) {
                    var _this = this;
                    var words = this.data.dataPoints;
                    if (_.isEmpty(words)) {
                        this.clear();
                        return;
                    }
                    requestAnimationFrame(function () {
                        var surface = _.range(0, (_this.specialViewport.width >> 5) * _this.specialViewport.height, 0);
                        words.forEach(function (data) {
                            data.getWidthOfWord = function () { return data.widthOfWord || (data.widthOfWord = powerbi.TextMeasurementService.measureSvgTextWidth({
                                fontFamily: _this.fontFamily,
                                fontSize: (data.size + 1) + WordCloud.Size,
                                //fontWeight: "normal",
                                //fontStyle: "normal",
                                text: data.text
                            }) + 2); };
                        });
                        _this.computeCycle(words, _this.getCanvasContext(), surface, null, onPositionsComputed, [], 0);
                    });
                };
                WordCloud.prototype.computeCycle = function (words, context, surface, borders, onPositionsComputed, wordsForDraw, index) {
                    if (wordsForDraw === void 0) { wordsForDraw = []; }
                    if (index === void 0) { index = 0; }
                    var word = words[index], ratio = 1;
                    if (words.length <= 10)
                        ratio = 5;
                    else if (words.length <= 25)
                        ratio = 3;
                    else if (words.length <= 75)
                        ratio = 1.5;
                    else if (words.length <= 100)
                        ratio = 1.25;
                    word.x = (this.specialViewport.width / ratio * (Math.random() + 0.5)) >> 1;
                    word.y = (this.specialViewport.height / ratio * (Math.random() + 0.5)) >> 1;
                    if (!word.sprite) {
                        this.generateSprites(context, words, index);
                    }
                    if (word.sprite && this.findPosition(surface, word, borders)) {
                        wordsForDraw.push(word);
                        borders = this.updateBorders(word, borders);
                        word.x -= this.specialViewport.width >> 1;
                        word.y -= this.specialViewport.height >> 1;
                        if (wordsForDraw.length >= this.settings.general.maxNumberOfWords) {
                            index = words.length - 1;
                        }
                    }
                    if (++index < words.length && this.root) {
                        this.computeCycle(words, context, surface, borders, onPositionsComputed, wordsForDraw, index);
                    }
                    else {
                        onPositionsComputed({
                            data: wordsForDraw,
                            leftBorder: borders && borders[0],
                            rightBorder: borders && borders[1]
                        });
                    }
                };
                WordCloud.prototype.updateBorders = function (word, borders) {
                    if (borders && borders.length === 2) {
                        var leftBorder = borders[0], rightBorder = borders[1];
                        if (word.x + word.x0 < leftBorder.x)
                            leftBorder.x = word.x + word.x0;
                        if (word.y + word.y0 < leftBorder.y)
                            leftBorder.y = word.y + word.y0;
                        if (word.x + word.x1 > rightBorder.x)
                            rightBorder.x = word.x + word.x1;
                        if (word.y + word.y1 > rightBorder.y)
                            rightBorder.y = word.y + word.y1;
                    }
                    else {
                        borders = [
                            {
                                x: word.x + word.x0,
                                y: word.y + word.y0
                            }, {
                                x: word.x + word.x1,
                                y: word.y + word.y1
                            }
                        ];
                    }
                    return borders;
                };
                WordCloud.prototype.generateSprites = function (context, words, startIndex) {
                    context.clearRect(0, 0, this.canvasViewport.width << 5, this.canvasViewport.height);
                    var x = 0, y = 0, maxHeight = 0;
                    for (var i = startIndex, length = words.length; i < length; i++) {
                        var currentWordData = words[i];
                        var widthOfWord = currentWordData.getWidthOfWord();
                        var heightOfWord = currentWordData.size << 1;
                        if (currentWordData.rotate) {
                            var sr = Math.sin(currentWordData.rotate * WordCloud.Radians), cr = Math.cos(currentWordData.rotate * WordCloud.Radians), widthCr = widthOfWord * cr, widthSr = widthOfWord * sr, heightCr = heightOfWord * cr, heightSr = heightOfWord * sr;
                            widthOfWord = (Math.max(Math.abs(widthCr + heightSr), Math.abs(widthCr - heightSr)) + 31) >> 5 << 5;
                            heightOfWord = Math.floor(Math.max(Math.abs(widthSr + heightCr), Math.abs(widthSr - heightCr)));
                        }
                        else {
                            widthOfWord = (widthOfWord + 31) >> 5 << 5;
                        }
                        if (heightOfWord > maxHeight) {
                            maxHeight = heightOfWord;
                        }
                        if (x + widthOfWord >= (this.canvasViewport.width << 5)) {
                            x = 0;
                            y += maxHeight;
                            maxHeight = 0;
                        }
                        context.save();
                        context.font = "normal normal " + (currentWordData.size + 1) + WordCloud.Size + " " + this.fontFamily;
                        context.translate((x + (widthOfWord >> 1)), (y + (heightOfWord >> 1)));
                        if (currentWordData.rotate) {
                            context.rotate(currentWordData.rotate * WordCloud.Radians);
                        }
                        context.fillText(currentWordData.text, 0, 0);
                        if (currentWordData.padding) {
                            context.lineWidth = 2 * currentWordData.padding;
                            context.strokeText(currentWordData.text, 0, 0);
                        }
                        context.restore();
                        currentWordData.width = widthOfWord;
                        currentWordData.height = heightOfWord;
                        currentWordData.xOff = x;
                        currentWordData.yOff = y;
                        currentWordData.x1 = widthOfWord >> 1;
                        currentWordData.y1 = heightOfWord >> 1;
                        currentWordData.x0 = -currentWordData.x1;
                        currentWordData.y0 = -currentWordData.y1;
                        x += widthOfWord;
                    }
                    this.setSprites(context, words);
                };
                WordCloud.prototype.setSprites = function (context, words) {
                    var pixels = context.getImageData(0, 0, this.canvasViewport.width << 5, this.canvasViewport.height).data;
                    var sprites = [];
                    for (var i = words.length - 1; i >= 0; i--) {
                        var currentWordData = words[i], width = currentWordData.width, width32 = width >> 5, height = currentWordData.y1 - currentWordData.y0, x = 0, y = 0, seen = 0, seenRow = 0;
                        if (currentWordData.xOff + width >= (this.canvasViewport.width << 5) ||
                            currentWordData.yOff + height >= this.canvasViewport.height) {
                            currentWordData.sprite = null;
                            continue;
                        }
                        for (var j = 0; j < height * width32; j++) {
                            sprites[j] = 0;
                        }
                        if (currentWordData.xOff !== null) {
                            x = currentWordData.xOff;
                        }
                        else {
                            return;
                        }
                        y = currentWordData.yOff;
                        seen = 0;
                        seenRow = -1;
                        for (var j = 0; j < height; j++) {
                            for (var k = 0; k < width; k++) {
                                var l = width32 * j + (k >> 5);
                                var index = ((y + j) * (this.canvasViewport.width << 5) + (x + k)) << 2;
                                var m = pixels[index]
                                    ? 1 << (31 - (k % 32))
                                    : 0;
                                sprites[l] |= m;
                                seen |= m;
                            }
                            if (seen) {
                                seenRow = j;
                            }
                            else {
                                currentWordData.y0++;
                                height--;
                                j--;
                                y++;
                            }
                        }
                        currentWordData.y1 = currentWordData.y0 + seenRow;
                        currentWordData.sprite = sprites.slice(0, (currentWordData.y1 - currentWordData.y0) * width32);
                    }
                };
                WordCloud.prototype.findPosition = function (surface, word, borders) {
                    var startPoint = { x: word.x, y: word.y }, delta = Math.sqrt(this.specialViewport.width * this.specialViewport.width + this.specialViewport.height * this.specialViewport.height), point, dt = Math.random() < 0.5 ? 1 : -1, shift = -dt, dx, dy;
                    while (true) {
                        shift += dt;
                        point = this.archimedeanSpiral(shift);
                        dx = Math.floor(point.x);
                        dy = Math.floor(point.y);
                        if (Math.min(Math.abs(dx), Math.abs(dy)) >= delta) {
                            break;
                        }
                        word.x = startPoint.x + dx;
                        word.y = startPoint.y + dy;
                        if (word.x + word.x0 < 0 ||
                            word.y + word.y0 < 0 ||
                            word.x + word.x1 > this.specialViewport.width ||
                            word.y + word.y1 > this.specialViewport.height)
                            continue;
                        if (!borders || !this.checkIntersect(word, surface)) {
                            if (!borders || this.checkIntersectOfRectangles(word, borders[0], borders[1])) {
                                var sprite = word.sprite, width = word.width >> 5, shiftWidth = this.specialViewport.width >> 5, lx = word.x - (width << 4), sx = lx & 127, msx = 32 - sx, height = word.y1 - word.y0, x = (word.y + word.y0) * shiftWidth + (lx >> 5);
                                for (var i = 0; i < height; i++) {
                                    var lastSprite = 0;
                                    for (var j = 0; j <= width; j++) {
                                        var leftMask = lastSprite << msx, rightMask;
                                        if (j < width)
                                            lastSprite = sprite[i * width + j];
                                        rightMask = j < width
                                            ? lastSprite >>> sx
                                            : 0;
                                        surface[x + j] |= leftMask | rightMask;
                                    }
                                    x += shiftWidth;
                                }
                                word.sprite = null;
                                return true;
                            }
                        }
                    }
                    return false;
                };
                WordCloud.prototype.archimedeanSpiral = function (value) {
                    var ratio = this.specialViewport.width / this.specialViewport.height;
                    value = value * 0.1;
                    return {
                        x: ratio * value * Math.cos(value),
                        y: value * Math.sin(value)
                    };
                };
                WordCloud.prototype.checkIntersect = function (word, surface) {
                    var shiftWidth = this.specialViewport.width >> 5, sprite = word.sprite, widthOfWord = word.width >> 5, lx = word.x - (widthOfWord << 4), sx = lx & 127, msx = 32 - sx, heightOfWord = word.y1 - word.y0, x = (word.y + word.y0) * shiftWidth + (lx >> 5);
                    for (var i = 0; i < heightOfWord; i++) {
                        var lastSprite = 0;
                        for (var j = 0; j <= widthOfWord; j++) {
                            var mask = 0, leftMask, intersectMask = 0;
                            leftMask = lastSprite << msx;
                            if (j < widthOfWord)
                                lastSprite = sprite[i * widthOfWord + j];
                            mask = j < widthOfWord
                                ? lastSprite >>> sx
                                : 0;
                            intersectMask = (leftMask | mask) & surface[x + j];
                            if (intersectMask)
                                return true;
                        }
                        x += shiftWidth;
                    }
                    return false;
                };
                WordCloud.prototype.checkIntersectOfRectangles = function (word, leftBorder, rightBorder) {
                    return (word.x + word.x1) > leftBorder.x &&
                        (word.x + word.x0) < rightBorder.x &&
                        (word.y + word.y1) > leftBorder.y &&
                        (word.y + word.y0) < rightBorder.y;
                };
                WordCloud.prototype.getCanvasContext = function () {
                    if (!this.canvasViewport)
                        return null;
                    this.canvas.width = 1;
                    this.canvas.height = 1;
                    var context = this.canvas.getContext("2d");
                    this.canvas.width = this.canvasViewport.width << 5;
                    this.canvas.height = this.canvasViewport.height;
                    context = this.canvas.getContext("2d");
                    context.fillStyle = context.strokeStyle = "red";
                    context.textAlign = "center";
                    return context;
                };
                WordCloud.prototype.UpdateSize = function () {
                    var fakeWidth, fakeHeight, ratio;
                    ratio = Math.sqrt((this.fakeViewport.width * this.fakeViewport.height)
                        / (this.layout.viewportIn.width * this.layout.viewportIn.height));
                    if (isNaN(ratio)) {
                        fakeHeight = fakeWidth = 1;
                    }
                    else {
                        fakeHeight = this.layout.viewportIn.height * ratio;
                        fakeWidth = this.layout.viewportIn.width * ratio;
                    }
                    this.specialViewport = {
                        height: fakeHeight,
                        width: fakeWidth
                    };
                    this.root.attr({
                        "height": this.layout.viewport.height,
                        "width": this.layout.viewport.width
                    });
                };
                WordCloud.prototype.render = function (wordCloudDataView) {
                    var _this = this;
                    if (!wordCloudDataView || !wordCloudDataView.data) {
                        return;
                    }
                    this.scaleMainView(wordCloudDataView);
                    this.wordsGroupUpdateSelection = this.main
                        .select(WordCloud.Words.selector)
                        .selectAll("g")
                        .data(wordCloudDataView.data);
                    var wordGroupEnterSelection = this.wordsGroupUpdateSelection
                        .enter()
                        .append("svg:g")
                        .classed(WordCloud.WordGroup.class, true);
                    wordGroupEnterSelection
                        .append("svg:text")
                        .style("font-size", "1px")
                        .attr('pointer-events', "none");
                    wordGroupEnterSelection
                        .append("svg:rect");
                    this.wordsGroupUpdateSelection.exit().remove();
                    this.wordsGroupUpdateSelection
                        .attr('transform', function (d) { return (visuals.SVGUtil.translate(d.x, d.y) + " rotate(" + d.rotate + ")"); })
                        .sort(function (a, b) { return b.height * b.width - a.height * a.width; });
                    this.wordsTextUpdateSelection = this.wordsGroupUpdateSelection.selectAll("text").data(function (d) { return [d]; });
                    this.wordsTextUpdateSelection.text(function (d) { return d.text; });
                    this.animation(this.wordsTextUpdateSelection, this.durationAnimations)
                        .style({
                        "font-size": (function (item) { return ("" + item.size + WordCloud.Size); }),
                        "fill": (function (item) { return item.color; }),
                    });
                    this.wordsGroupUpdateSelection.selectAll("rect").data(function (d) { return [d]; })
                        .attr({
                        x: function (d) { return -d.getWidthOfWord() * 0.5; },
                        width: function (d) { return d.getWidthOfWord(); },
                        y: function (d) { return -d.size * 0.75; },
                        height: function (d) { return d.size * 0.85; },
                        fill: function (d) { return "rgba(63, 191, 191, 0.0)"; },
                    })
                        .on("click", function (d) { _this.setSelection(d); d3.event.stopPropagation(); });
                    this.clearIntorrectSelection(this.data.dataView);
                    this.renderSelection();
                    this.isUpdating = false;
                    if (this.incomingUpdateOptions !== this.visualUpdateOptions) {
                        this.update(this.incomingUpdateOptions);
                    }
                };
                WordCloud.prototype.clearIntorrectSelection = function (dataView) {
                    var _this = this;
                    var categories = dataView && dataView.categorical && dataView.categorical.categories;
                    var identityKeys = categories && categories[0] && categories[0].identity && categories[0].identity.map(function (x) { return x.key; });
                    var oldIdentityKeys = this.oldIdentityKeys;
                    this.oldIdentityKeys = identityKeys;
                    if (oldIdentityKeys && oldIdentityKeys.length > identityKeys.length) {
                        this.selectionManager.clear(false);
                        return;
                    }
                    if (!_.isEmpty(identityKeys)) {
                        var incorrectValues = this.selectionManager.getSelectionIdValues.filter(function (x) {
                            return x.selectionId.some(function (s) { return _.contains(identityKeys, s.getKey()); });
                        });
                        incorrectValues.forEach(function (v) { return _this.selectionManager.selectedValues
                            .splice(_this.selectionManager.selectedValues.indexOf(v.value), 1); });
                    }
                };
                WordCloud.prototype.setSelection = function (dataPoint) {
                    var _this = this;
                    if (!dataPoint) {
                        this.selectionManager.clear(true).then(function () { return _this.renderSelection(); });
                        return;
                    }
                    this.selectionManager
                        .selectAndSendSelection(dataPoint.text, d3.event.ctrlKey)
                        .then(function () { return _this.renderSelection(); });
                };
                WordCloud.prototype.scaleMainView = function (wordCloudDataView) {
                    var rectangles = wordCloudDataView.data.map(function (d) {
                        var hw = d.width / 2;
                        var hh = d.height / 2;
                        return { left: d.x - hw, top: d.y - hh, right: d.x + hw, bottom: d.y + hh };
                    });
                    if (_.isEmpty(rectangles)) {
                        return;
                    }
                    var rectangle = {
                        left: _.min(rectangles, function (x) { return x.left; }).left,
                        top: _.min(rectangles, function (x) { return x.top; }).top,
                        right: _.max(rectangles, function (x) { return x.right; }).right,
                        bottom: _.max(rectangles, function (x) { return x.bottom; }).bottom
                    };
                    rectangle.width = rectangle.right - rectangle.left;
                    rectangle.height = rectangle.bottom - rectangle.top;
                    var scaleByX = this.layout.viewportIn.width / (rectangle.width);
                    var scaleByY = this.layout.viewportIn.height / (rectangle.height);
                    var scale = Math.min(scaleByX, scaleByY);
                    var x = -rectangle.left * scale + 5;
                    var y = -rectangle.top * scale + 5;
                    this.main
                        .style("line-height", "5px"); //TODO: This construction fixes bug #6343.
                    this.main
                        .attr("transform", visuals.SVGUtil.translate(x, y) + " scale(" + scale + ")")
                        .style("line-height", "10px"); //TODO: This construction fixes bug #6343.
                };
                WordCloud.prototype.renderSelection = function () {
                    var _this = this;
                    if (!this.wordsTextUpdateSelection) {
                        return;
                    }
                    if (!this.selectionManager.hasSelection) {
                        this.setOpacity(this.wordsTextUpdateSelection, WordCloud.MaxOpacity);
                        return;
                    }
                    var selectedColumns = this.wordsTextUpdateSelection.filter(function (d) {
                        return _this.selectionManager.isSelected(d.text);
                    });
                    this.setOpacity(this.wordsTextUpdateSelection, WordCloud.MinOpacity);
                    this.setOpacity(selectedColumns, WordCloud.MaxOpacity);
                };
                WordCloud.prototype.setOpacity = function (element, opacityValue) {
                    element.style("fill-opacity", opacityValue);
                    if (this.main) {
                        this.main.style("line-height", "14px");
                        this.animation(this.main, 0, this.durationAnimations)
                            .style("line-height", "15px");
                    }
                };
                WordCloud.prototype.enumerateObjectInstances = function (options) {
                    var _this = this;
                    var instances = WordCloudSettings.enumerateObjectInstances(this.settings && this.settings.originalSettings, options, WordCloud.capabilities);
                    switch (options.objectName) {
                        case "dataPoint":
                            if (this.data && this.data.dataPoints) {
                                var wordCategoriesIndex = [];
                                _.unique(this.data.dataPoints, function (x) { return x.wordIndex; }).forEach(function (item) {
                                    if (wordCategoriesIndex.indexOf(item.wordIndex) === -1) {
                                        wordCategoriesIndex.push(item.wordIndex);
                                        instances.pushInstance({
                                            objectName: options.objectName,
                                            displayName: _this.data.texts[item.wordIndex].text,
                                            selector: visuals.ColorHelper.normalizeSelector(item.selectionIds[0].getSelector(), false),
                                            properties: { fill: { solid: { color: item.color } } }
                                        });
                                    }
                                });
                            }
                            break;
                    }
                    return instances.complete();
                };
                WordCloud.prototype.animation = function (element, duration, delay, callback) {
                    if (duration === void 0) { duration = 0; }
                    if (delay === void 0) { delay = 0; }
                    return element
                        .transition()
                        .delay(delay)
                        .duration(duration)
                        .each("end", callback);
                };
                WordCloud.prototype.destroy = function () {
                    this.root = null;
                    this.canvas = null;
                };
                WordCloud.ClassName = "wordCloud";
                WordCloud.Words = {
                    "class": "words",
                    selector: ".words"
                };
                WordCloud.WordGroup = {
                    "class": "word",
                    selector: ".word"
                };
                WordCloud.Size = "px";
                WordCloud.StopWordsDelemiter = " ";
                WordCloud.Radians = Math.PI / 180;
                WordCloud.MinAngle = -180;
                WordCloud.MaxAngle = 180;
                WordCloud.MaxNumberOfWords = 2500;
                WordCloud.MinOpacity = 0.2;
                WordCloud.MaxOpacity = 1;
                WordCloud.FontSizePercentageCoefficent = 1;
                WordCloud.capabilities = {
                    dataRoles: [{
                            name: WordCloudColumns.Roles.Category,
                            kind: powerbi.VisualDataRoleKind.Grouping,
                            displayName: "Category"
                        }, {
                            name: WordCloudColumns.Roles.Values,
                            kind: powerbi.VisualDataRoleKind.Measure,
                            displayName: "Values"
                        }],
                    dataViewMappings: [{
                            conditions: [{
                                    "Category": {
                                        min: 1,
                                        max: 1
                                    },
                                    "Values": {
                                        min: 0,
                                        max: 1
                                    }
                                }],
                            categorical: {
                                categories: {
                                    for: { in: "Category" },
                                    dataReductionAlgorithm: { top: { count: WordCloud.MaxNumberOfWords } }
                                },
                                values: {
                                    for: { in: "Values" }
                                }
                            }
                        }],
                    sorting: {
                        implicit: {
                            clauses: [{
                                    role: "Values",
                                    direction: 2 /*SortDirection.Descending*/ //Constant SortDirection.Descending currently is not supported on the msit
                                }]
                        }
                    },
                    objects: {
                        general: {
                            displayName: "General",
                            properties: {
                                formatString: {
                                    type: {
                                        formatting: {
                                            formatString: true
                                        }
                                    }
                                },
                                maxNumberOfWords: {
                                    displayName: "Max number of words",
                                    type: { numeric: true }
                                },
                                minFontSize: {
                                    displayName: "Min Font",
                                    type: { formatting: { fontSize: true } }
                                },
                                maxFontSize: {
                                    displayName: "Max Font",
                                    type: { formatting: { fontSize: true } }
                                },
                                isBrokenText: {
                                    displayName: "Word-breaking",
                                    type: { bool: true }
                                },
                            }
                        },
                        dataPoint: {
                            displayName: "Data colors",
                            properties: {
                                fill: {
                                    displayName: "Fill",
                                    type: { fill: { solid: { color: true } } }
                                }
                            }
                        },
                        stopWords: {
                            displayName: "Stop Words",
                            properties: {
                                show: {
                                    displayName: "Show",
                                    type: { bool: true }
                                },
                                isDefaultStopWords: {
                                    displayName: "Default Stop Words",
                                    type: { bool: true }
                                },
                                words: {
                                    displayName: "Words",
                                    type: { text: true }
                                }
                            }
                        },
                        rotateText: {
                            displayName: "Rotate Text",
                            properties: {
                                show: {
                                    displayName: "Show",
                                    type: { bool: true }
                                },
                                minAngle: {
                                    displayName: "Min Angle",
                                    type: { numeric: true }
                                },
                                maxAngle: {
                                    displayName: "Max Angle",
                                    type: { numeric: true }
                                },
                                maxNumberOfOrientations: {
                                    displayName: "Max number of orientations",
                                    type: { numeric: true }
                                }
                            }
                        }
                    }
                };
                WordCloud.Punctuation = [
                    "!", ".", ":", "'", ";", ",", "!",
                    "@", "#", "$", "%", "^", "&", "*",
                    "(", ")", "[", "]", "\"", "\\", "/",
                    "-", "_", "+", "="
                ];
                WordCloud.StopWords = [
                    "a", "able", "about", "across", "after", "all", "almost", "also", "am", "among", "an",
                    "and", "any", "are", "as", "at", "be", "because", "been", "but", "by", "can", "cannot",
                    "could", "did", "do", "does", "either", "else", "ever", "every", "for", "from", "get",
                    "got", "had", "has", "have", "he", "her", "hers", "him", "his", "how", "however", "i",
                    "if", "in", "into", "is", "it", "its", "just", "least", "let", "like", "likely", "may",
                    "me", "might", "most", "must", "my", "neither", "no", "nor", "not", "of", "off", "often",
                    "on", "only", "or", "other", "our", "own", "rather", "said", "say", "says", "she", "should",
                    "since", "so", "some", "than", "that", "the", "their", "them", "then", "there", "these",
                    "they", "this", "tis", "to", "too", "twas", "us", "wants", "was", "we", "were", "what",
                    "when", "where", "which", "while", "who", "whom", "why", "will", "with", "would", "yet",
                    "you", "your"
                ];
                WordCloud.DefaultMargin = {
                    top: 10,
                    right: 10,
                    bottom: 10,
                    left: 10
                };
                return WordCloud;
            }());
            WordCloud1447959067750.WordCloud = WordCloud;
            var explore;
            (function (explore) {
                var util;
                (function (util) {
                    function hexToRgb(hex) {
                        // Expand shorthand form (e.g. "03F") to full form (e.g. "0033FF")
                        var shorthandRegex = /^#?([a-f\d])([a-f\d])([a-f\d])$/i;
                        hex = hex.replace(shorthandRegex, function (m, r, g, b) {
                            return r + r + g + g + b + b;
                        });
                        var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
                        return result ? "rgb(" + parseInt(result[1], 16) + "," + parseInt(result[2], 16) + "," + parseInt(result[3], 16) + ")" : null;
                    }
                    util.hexToRgb = hexToRgb;
                    function getRandomColor() {
                        var red = Math.floor(Math.random() * 255), green = Math.floor(Math.random() * 255), blue = Math.floor(Math.random() * 255);
                        return "rgb(" + red + "," + green + "," + blue + ")";
                    }
                    util.getRandomColor = getRandomColor;
                })(util = explore.util || (explore.util = {}));
            })(explore || (explore = {}));
        })(WordCloud1447959067750 = visuals.WordCloud1447959067750 || (visuals.WordCloud1447959067750 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.WordCloud1447959067750 = {
                name: 'WordCloud1447959067750',
                class: 'WordCloud1447959067750',
                capabilities: powerbi.visuals.WordCloud1447959067750.WordCloud.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.WordCloud1447959067750.WordCloud(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
